/**
	This file is part of Hedge Maze.

    Hedge Maze is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hedge Maze is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hedge Maze.  If not, see <http://www.gnu.org/licenses/>.

	Copyright 2006, 2010 James Hill
*/

#include "maze.h"

// the default constructor will set the width and height to 10
maze::maze():mazeHeight(10),mazeWidth(10)
{
	theMaze = new mazeNode [mazeWidth*mazeHeight];
}

// the overloaded constructor will take a height and a width and 
// initialize the maze to that
maze::maze(unsigned char width, unsigned char height)
{
	// set the maze width and height
	mazeWidth = width;
	mazeHeight = height;

	theMaze = new mazeNode [mazeWidth * mazeHeight];
}

// the copy constructor will alocate new memory for all of the 
// nodes and copy over the width and height
maze::maze(const maze& toCopy)
{
	mazeWidth = toCopy.getWidth();
	mazeHeight = toCopy.getHeight();

	// dealocate the memory if it's being used
	delete [] theMaze;

	// reallocate
	theMaze = new mazeNode [mazeWidth * mazeHeight];

	// copy the values at each of the nodes to the new location
	for(char i=0; i < mazeWidth; i++)
		for(char j=0; j < mazeHeight; j++)
			this->getNode(i,j)->node=toCopy.getNode(i,j)->node;
}

// the destructor simply deallocates the memory for the 
// maze nodes
maze::~maze()
{
	// deallocate the memory
	delete [] theMaze;
}

// returns a pointer to the give maze node
// allows for indexing like a multi-dimensional array
mazeNode * maze::getNode(unsigned char x, unsigned char y) const
{
	return &theMaze[y*mazeHeight+x];
}

// returns the width and height of th maze
unsigned char maze::getWidth() const
{
	return mazeWidth;
}

unsigned char maze::getHeight() const
{
	return mazeHeight;
}

// returns a maze to the array of mazeNodes
mazeNode * maze::getMaze()
{
	return theMaze;
}

// sets the width and height of the maze
void maze::setWidth(unsigned char newWidth)
{
	mazeWidth=newWidth;
}

void maze::setHeight(unsigned char newHeight)
{
	mazeHeight=newHeight;
}

// The recursive funciton that actually creats the maze
void maze::createMaze(unsigned char nx, unsigned char ny, unsigned char entry)
{
	unsigned char nextPath;

	// flag that this node has been reached by setting on of its
	// entry point variables
	switch (entry) {
		case 0: // entered from the top of the last node
			getNode(nx,ny)->epbot = true;
			break;
		case 1: // entered from the bottom of the last node
			getNode(nx,ny)->eptop = true;
			break;
		case 2: // entered from the right of the last node
			getNode(nx,ny)->epleft = true;
			break;
		case 3: // entered from the left of the last node
			getNode(nx,ny)->epright = true;
			break;
	}

	// check if in one of the corners
	// if so, return
	if(nx == 0 && (ny == 0 || ny == mazeHeight-1))
		return;
	else if(nx == mazeWidth-1 && (ny == 0 || ny == mazeHeight-1))
		return;

	// otherwise call the function on the next path
	do {
		nextPath = choosePath(nx,ny);
		switch (nextPath) {
			case 0: // top
				getNode(nx,ny)->eptop = true;
				createMaze(nx,ny+1,0); // top node, entering from the top
				break;
			case 1: // bot
				getNode(nx,ny)->epbot = true;
				createMaze(nx,ny-1,1); // bot node, entering from the bottom
				break;
			case 2: // right
				getNode(nx,ny)->epright = true;
				createMaze(nx+1,ny,2); // right node, entering form the right
				break;
			case 3: // left
				getNode(nx,ny)->epleft = true;
				createMaze(nx-1,ny,3); // left node, entering from left
		} 
	}while(nextPath!=5);
}

// check to see if node is available
bool maze::isAvailable (unsigned char nx, unsigned char ny)
{
	return (!getNode(nx,ny)->eptop &&
			!getNode(nx,ny)->epbot &&
			!getNode(nx,ny)->epright &&
			!getNode(nx,ny)->epleft);
}

// chooses a random path to follow
unsigned char maze::choosePath (unsigned char nx, unsigned char ny)
{
	unsigned int pick;

	// check to see if available
	bool isSomethingAvailable = false;
	if(ny+1 != mazeHeight && isAvailable(nx,ny+1)) // top
		isSomethingAvailable = true;
		//return (0);
	if(ny != 0 && isAvailable(nx,ny-1)) // bot
		isSomethingAvailable = true;
		//return (1);
	if(nx+1 != mazeWidth && isAvailable(nx+1,ny)) // right
		isSomethingAvailable = true;
		//return (2);
	if(nx != 0 && isAvailable(nx-1,ny)) // left
		isSomethingAvailable = true;
		//return (3);
	
	if(!isSomethingAvailable)
		return (5);

	while(true) {
		pick = rand();

		if(pick%4 == 0)
			if(ny+1 != mazeHeight && isAvailable(nx,ny+1)) // top
				return (0);
		if(pick%4 == 1)
			if(ny != 0 && isAvailable(nx,ny-1)) // bot
				return (1);
		if(pick%4 == 2)
			if(nx+1 != mazeWidth && isAvailable(nx+1,ny)) // right
				return (2);
		if(pick%4 == 3)
			if(nx != 0 && isAvailable(nx-1,ny)) // left
				return (3);
	}
}

// initialize the maze
void maze::initMaze()
{
	unsigned char w, h;
	unsigned char ex;	// entry coordinates

	// 1. initialize all walls in the maze to false except for the boundries
	for(w = 0; w < mazeWidth; w++) {
		for(h = 0; h < mazeHeight; h++) {
			getNode(w,h)->node=0;	// initialize all walls to false;

			// if the width is either 0 or the max, than set the respective wall to zero
			if		(w == 0) {
				getNode(0,h)->left = true; 
			}
			else if (w == mazeWidth-1) {
				getNode(mazeWidth-1,h)->right = true;
			}
			
			// do the same for the height
			if		(h == 0) {
				getNode(w,0)->bot = true;
			}
			else if (h == mazeHeight-1) {
				getNode(w,mazeHeight-1)->top = true;
			}
		}
	}

	// 2. choose and entry and an exit node (mWidth/2)
	ex=mazeWidth/2;

	srand((unsigned)time(NULL));

	// 3. call mazeGen to generate the maze
	createMaze(ex,0,0);  // starting at width/2 0 and entering from the top of the last node
}

// fills in the walls of the maze
void maze::fillWalls()
{
	int i,j;
	for(i=0;i<mazeWidth;i++) {
		for(j=0;j<mazeHeight;j++) {
			if(!getNode(i,j)->eptop)
				getNode(i,j)->top = true;
			else
				getNode(i,j)->top = false;
			if(!getNode(i,j)->epbot)
				getNode(i,j)->bot = true;
			else
				getNode(i,j)->bot = false;
			if(!getNode(i,j)->epright)
				getNode(i,j)->right = true;
			else
				getNode(i,j)->right = false;
			if(!getNode(i,j)->epleft)
				getNode(i,j)->left = true;
			else
				getNode(i,j)->left = false;
		}
	}

	// 4. open the exit
	getNode(0,mazeHeight-1)->top = false;
}

// draws an ascii version of the maze to a file named maze
void maze::outputASCIImaze()
{
	fillWalls();

	std::ofstream mazeOut("maze.txt");

	for(int i = mazeHeight; i > 0; i--) {
		for (int k = 0; k < 2; k++) {
			for (int j = 0; j < mazeWidth; j++) {
				if(k == 0) {
					if( getNode(j,i-1)->top == true && j != 0)
						mazeOut << "----";
					else if( getNode(j, i-1)->top == true && j == 0)
						mazeOut << "|---";
					else if( getNode(j, i-1)->top == false && j == 0)
						mazeOut << "|   ";
					else 
						mazeOut << "    ";
				}
				if(k == 1) {
					if( getNode(j,i-1)->left == true )
						mazeOut << "|";
					else 
						mazeOut << " ";
					mazeOut << "   ";
				}
				if( j == mazeWidth - 1)
					mazeOut << "|";
			}
			mazeOut << std::endl;
		}
	}
	for(int i = 0; i < mazeWidth; i++)
		mazeOut << "----";

	mazeOut.close();
}